#ifndef AddImpl_ref_cpp_H_
#define AddImpl_ref_cpp_H_

#include "utils/Types.h"

#include "backend/OperatorImpl.hpp"
#include "operator/Add.hpp"
#include "utils/Registrar.hpp"
#include <memory>
#include <vector>

namespace Aidge {
// class Add_Op<2>;

// compute kernel registry for forward and backward
template <DimIdx_t NUM>
class AddImplForward_ref_cpp;
template <DimIdx_t NUM>
class AddImplBackward_ref_cpp;

template <>
class AddImplForward_ref_cpp<1>
    : public Registrable<std::tuple<DataType, DataType>, void(const std::size_t, const void*, void*)> {};
template <>
class AddImplBackward_ref_cpp<1>
    : public Registrable<std::tuple<DataType, DataType>, void(const std::size_t, const void*, void*)> {};

template <>
class AddImplForward_ref_cpp<2> : public Registrable<std::tuple<DataType, DataType, DataType>,
                                                     void(const std::size_t, const void*, const void*, void*)> {};
template <>
class AddImplBackward_ref_cpp<2> : public Registrable<std::tuple<DataType, DataType, DataType>,
                                                      void(const std::size_t, const void*, const void*, void*)> {};

template <>
class AddImplForward_ref_cpp<3> : public Registrable<std::tuple<DataType, DataType, DataType, DataType>,
                                                     void(const std::size_t, const void*, const void*, const void*, void*)> {
};
template <>
class AddImplBackward_ref_cpp<3>
    : public Registrable<std::tuple<DataType, DataType, DataType, DataType>,
                         void(const std::size_t, const void*, const void*, const void*, void*)> {};

template <DimIdx_t NUM>
class AddImpl_ref_cpp : public OperatorImpl {
   private:
    const Add_Op<NUM>& mOp;
    std::array<NbElts_t, NUM> mNbConsumedData = {};
    std::array<NbElts_t, 1> mNbProducedData = {};

   public:
    AddImpl_ref_cpp(const Add_Op<NUM>& op) : mOp(op) {}

    static std::unique_ptr<AddImpl_ref_cpp<NUM>> create(const Add_Op<NUM>& op) {
        return std::make_unique<AddImpl_ref_cpp<NUM>>(op);
    }

   public:
    NbElts_t getNbRequiredData(IOIndex_t inputIdx) const override final {
        assert(mOp.getInput(inputIdx) && "requires valid input");

        // Requires the whole tensors
        const auto& inputDims = std::static_pointer_cast<Tensor>(mOp.getInput(inputIdx))->dims();
        return std::accumulate(inputDims.begin(), inputDims.end(), NbElts_t(1), std::multiplies<NbElts_t>());
    }

    NbElts_t getNbRequiredProtected(IOIndex_t inputIdx) const override final {
        // for the direct convolution algorithm, convolutions can be in-place, if there is no padding!
        return 0;
    }

    NbElts_t getRequiredMemory(IOIndex_t outputIdx, const std::vector<DimSize_t>& inputsSize) const override final {
        // Requires the whole tensors, regardless of available data on inputs
        assert(outputIdx == 0 && "operator has only one output");

        const auto& outputDims = std::static_pointer_cast<Tensor>(mOp.getOutput(0))->dims();
        return std::accumulate(outputDims.begin(), outputDims.end(), NbElts_t(1), std::multiplies<NbElts_t>());
    }

    NbElts_t getNbConsumedData(IOIndex_t inputIdx) const override final {
        assert(inputIdx < mNbConsumedData.size());
        return mNbConsumedData[inputIdx];
    }

    NbElts_t getNbProducedData(IOIndex_t outputIdx) const override final {
        assert(outputIdx < mNbProducedData.size());
        return mNbProducedData[outputIdx];
    }

    void forward() {
        // nothing
    }

    void backward() { printf("Not implemented yet.\n"); }
};

template <>
class AddImpl_ref_cpp<1> : public OperatorImpl {
   private:
    const Add_Op<1>& mOp;
    std::array<NbElts_t, 1> mNbConsumedData;
    std::array<NbElts_t, 1> mNbProducedData;

   public:
    AddImpl_ref_cpp(const Add_Op<1>& op) : mOp(op), mNbConsumedData({0}), mNbProducedData({0}) {}

    static std::unique_ptr<AddImpl_ref_cpp<1>> create(const Add_Op<1>& op) {
        return std::make_unique<AddImpl_ref_cpp<1>>(op);
    }

   public:
    NbElts_t getNbRequiredData(IOIndex_t /*inputIdx*/) const override final;

    NbElts_t getNbRequiredProtected(IOIndex_t /*inputIdx*/) const override final;

    NbElts_t getRequiredMemory(IOIndex_t /*outputIdx*/,
                               const std::vector<DimSize_t>& /*inputsSize*/) const override final;

    NbElts_t getNbConsumedData(IOIndex_t /*inputIdx*/) const override final;

    NbElts_t getNbProducedData(IOIndex_t /*outputIdx*/) const override final;

    void forward();

    void backward();
};

template <>
class AddImpl_ref_cpp<2> : public OperatorImpl {
   private:
    const Add_Op<2>& mOp;
    std::array<NbElts_t, 2> mNbConsumedData;
    std::array<NbElts_t, 1> mNbProducedData;

   public:
    AddImpl_ref_cpp(const Add_Op<2>& op) : mOp(op), mNbConsumedData({0, 0}), mNbProducedData({0}) {}

    static std::unique_ptr<AddImpl_ref_cpp<2>> create(const Add_Op<2>& op) {
        return std::make_unique<AddImpl_ref_cpp<2>>(op);
    }

   public:
    NbElts_t getNbRequiredData(IOIndex_t inputIdx) const override final;

    NbElts_t getNbRequiredProtected(IOIndex_t inputIdx) const override final;

    NbElts_t getRequiredMemory(IOIndex_t /*outputIdx*/,
                               const std::vector<DimSize_t>& /*inputsSize*/) const override final;

    NbElts_t getNbConsumedData(IOIndex_t inputIdx) const override final;

    NbElts_t getNbProducedData(IOIndex_t /*outputIdx*/) const override final;

    void forward();

    void backward();
};

template <>
class AddImpl_ref_cpp<3> : public OperatorImpl {
   private:
    const Add_Op<3>& mOp;
    std::array<NbElts_t, 3> mNbConsumedData;
    std::array<NbElts_t, 1> mNbProducedData;

   public:
    AddImpl_ref_cpp(const Add_Op<3>& op) : mOp(op), mNbConsumedData({0, 0, 0}), mNbProducedData({0}) {}

    static std::unique_ptr<AddImpl_ref_cpp<3>> create(const Add_Op<3>& op) {
        return std::make_unique<AddImpl_ref_cpp<3>>(op);
    }

   public:
    NbElts_t getNbRequiredData(IOIndex_t inputIdx) const override final;

    NbElts_t getNbRequiredProtected(IOIndex_t /*inputIdx*/) const override final;

    NbElts_t getRequiredMemory(IOIndex_t outputIdx, const std::vector<DimSize_t>& /*inputsSize*/) const override final;

    NbElts_t getNbConsumedData(IOIndex_t inputIdx) const override final;

    NbElts_t getNbProducedData(IOIndex_t outputIdx) const override final;

    void forward();

    void backward();
};

namespace {
static Registrar<Add_Op<1>> registrarAddImpl1I_ref_cpp("cpu", Aidge::AddImpl_ref_cpp<1>::create);
static Registrar<Add_Op<2>> registrarAddImpl2I_ref_cpp("cpu", Aidge::AddImpl_ref_cpp<2>::create);
static Registrar<Add_Op<3>> registrarAddImpl3I_ref_cpp("cpu", Aidge::AddImpl_ref_cpp<3>::create);
}  // namespace
}  // namespace Aidge

#endif /* AddImpl_ref_cpp_H_ */