
#include <cassert>
#include <numeric>
#include <chrono>
#include <thread>
#include <vector>

#include "operator/Conv.hpp"

#include "operator/AddImpl.hpp"
#include "operator/AddImpl_kernels.hpp"
#include "utils/Types.h"

//////////////////////////////////
// AddImpl_ref_cpp<1>
//////////////////////////////////

Aidge::NbElts_t Aidge::AddImpl_ref_cpp<1>::getNbRequiredData(Aidge::IOIndex_t /*inputIdx*/) const {
    assert(mOp.getInput(0) && "requires valid input");
    // Requires the whole tensors
    return static_cast<int>(std::static_pointer_cast<Tensor>(mOp.getInput(0))->size());
}

Aidge::NbElts_t Aidge::AddImpl_ref_cpp<1>::getNbRequiredProtected(Aidge::IOIndex_t /*inputIdx*/) const {
    // for the direct convolution algorithm, convolutions can be in-place, if there is no padding!
    return 0;
}

Aidge::NbElts_t Aidge::AddImpl_ref_cpp<1>::getRequiredMemory(Aidge::IOIndex_t /*outputIdx*/, const std::vector<Aidge::DimSize_t>& /*inputsSize*/) const {
    // Requires the whole tensors, regardless of available data on inputs
    return std::static_pointer_cast<Tensor>(mOp.getOutput(0))->size();
}

Aidge::NbElts_t Aidge::AddImpl_ref_cpp<1>::getNbConsumedData(Aidge::IOIndex_t /*inputIdx*/) const {
    return mNbConsumedData[0];
}

Aidge::NbElts_t Aidge::AddImpl_ref_cpp<1>::getNbProducedData(Aidge::IOIndex_t /*outputIdx*/) const {
    return mNbProducedData[0];
}

void Aidge::AddImpl_ref_cpp<1>::forward() {
    // FIXME: uncomment the following code once memory handling will work
    assert(mOp.mInputs[0] && "missing input #0");

    // Find the correct kernel type
    auto kernelFunc = Registrar<AddImplForward_ref_cpp<1>>::create({
        mOp.mInputs[0]->dataType(),
        mOp.mOutput->dataType()});

    // Call kernel
    kernelFunc(std::static_pointer_cast<Tensor>(mOp.mInputs[0])->size(),
        mOp.mInputs[0]->getImpl()->rawPtr(),
        mOp.mOutput->getImpl()->rawPtr());

    // Update producer-consumer data
    for (IOIndex_t inputIdx = 0; static_cast<NbElts_t>(inputIdx) < mNbConsumedData.size(); ++inputIdx)
        mNbConsumedData[inputIdx]+= getNbRequiredData(inputIdx); // each input is consumed by the minimum amount for a forward pass

    mNbProducedData[0]+= getRequiredMemory(0, {});
}

void Aidge::AddImpl_ref_cpp<1>::backward() {
    printf("Not implemented yet.\n");
}


//////////////////////////////////
// AddImpl_ref_cpp<2>
//////////////////////////////////


Aidge::NbElts_t Aidge::AddImpl_ref_cpp<2>::getNbRequiredData(Aidge::IOIndex_t inputIdx) const {
    assert(mOp.getInput(inputIdx) && "requires valid input");

    // Requires the whole tensors
    const auto& inputDims = std::static_pointer_cast<Tensor>(mOp.getInput(inputIdx))->dims();

    return std::accumulate(inputDims.begin(), inputDims.end(),
                            NbElts_t(1), std::multiplies<NbElts_t>());
}

Aidge::NbElts_t Aidge::AddImpl_ref_cpp<2>::getNbRequiredProtected(Aidge::IOIndex_t /*inputIdx*/) const {
    // for the direct convolution algorithm, convolutions can be in-place, if there is no padding!
    return 0;
}

Aidge::NbElts_t Aidge::AddImpl_ref_cpp<2>::getRequiredMemory(Aidge::IOIndex_t outputIdx, const std::vector<Aidge::DimSize_t>& /*inputsSize*/) const {
    // Requires the whole tensors, regardless of available data on inputs
    assert(outputIdx == 0 && "operator has only one output");

    const auto& outputDims = std::static_pointer_cast<Tensor>(mOp.getOutput(0))->dims();
    return std::accumulate(outputDims.begin(), outputDims.end(),
                        NbElts_t(1), std::multiplies<NbElts_t>());
}

Aidge::NbElts_t Aidge::AddImpl_ref_cpp<2>::getNbConsumedData(Aidge::IOIndex_t inputIdx) const {
    assert(static_cast<std::size_t>(inputIdx) < mNbConsumedData.size());
    return mNbConsumedData[static_cast<std::size_t>(inputIdx)];
}

Aidge::NbElts_t Aidge::AddImpl_ref_cpp<2>::getNbProducedData(Aidge::IOIndex_t /*outputIdx*/) const {
    return mNbProducedData[0];
}

void Aidge::AddImpl_ref_cpp<2>::forward() {
    // FIXME: uncomment the following code once memory handling will work
    assert(mOp.mInputs[0] && "missing input #0");
    assert(mOp.mInputs[1] && "missing input #1");

    // Find the correct kernel type
    auto kernelFunc = Registrar<AddImplForward_ref_cpp<2>>::create({
        mOp.mInputs[0]->dataType(),
        mOp.mInputs[1]->dataType(),
        mOp.mOutput->dataType()});

    // Call kernel
    kernelFunc(std::static_pointer_cast<Tensor>(mOp.mInputs[0])->size(),
        mOp.mInputs[0]->getImpl()->rawPtr(),
        mOp.mInputs[1]->getImpl()->rawPtr(),
        mOp.mOutput->getImpl()->rawPtr());

    // Update producer-consumer data
    for (std::size_t inputIdx = 0; inputIdx < mNbConsumedData.size(); ++inputIdx)
        mNbConsumedData[inputIdx]+= getNbRequiredData(static_cast<IOIndex_t>(inputIdx)); // each input is consumed by the minimum amount for a forward pass

    mNbProducedData[0]+= getRequiredMemory(0, {});
}

void Aidge::AddImpl_ref_cpp<2>::backward() {
    printf("Not implemented yet.\n");
}


//////////////////////////////////
// AddImpl_ref_cpp<3>
//////////////////////////////////


Aidge::NbElts_t Aidge::AddImpl_ref_cpp<3>::getNbRequiredData(Aidge::IOIndex_t inputIdx) const {
    assert(mOp.getInput(inputIdx) && "requires valid input");

    // Requires the whole tensors
    const auto& inputDims = std::static_pointer_cast<Tensor>(mOp.getInput(inputIdx))->dims();

    return std::accumulate(inputDims.begin(), inputDims.end(),
                            Aidge::NbElts_t(1), std::multiplies<Aidge::NbElts_t>());
}

Aidge::NbElts_t Aidge::AddImpl_ref_cpp<3>::getNbRequiredProtected(Aidge::IOIndex_t /*inputIdx*/) const {
    // for the direct convolution algorithm, convolutions can be in-place, if there is no padding!
    return 0;
}

Aidge::NbElts_t Aidge::AddImpl_ref_cpp<3>::getRequiredMemory(Aidge::IOIndex_t outputIdx, const std::vector<Aidge::DimSize_t>& /*inputsSize*/) const {
    // Requires the whole tensors, regardless of available data on inputs
    assert(outputIdx == 0 && "operator has only one output");

    const auto& outputDims = std::static_pointer_cast<Tensor>(mOp.getOutput(0))->dims();
    return std::accumulate(outputDims.begin(), outputDims.end(),
                        NbElts_t(1), std::multiplies<NbElts_t>());
}

Aidge::NbElts_t Aidge::AddImpl_ref_cpp<3>::getNbConsumedData(Aidge::IOIndex_t inputIdx) const {
    assert(static_cast<std::size_t>(inputIdx) < mNbConsumedData.size());
    return mNbConsumedData[inputIdx];
}

Aidge::NbElts_t Aidge::AddImpl_ref_cpp<3>::getNbProducedData(Aidge::IOIndex_t outputIdx) const {
    assert(static_cast<std::size_t>(outputIdx) < mNbProducedData.size());
    return mNbProducedData[static_cast<std::size_t>(outputIdx)];
}

void Aidge::AddImpl_ref_cpp<3>::forward() {
    // FIXME: uncomment the following code once memory handling will work
    assert(mOp.mInputs[0] && "missing input #0");
    assert(mOp.mInputs[1] && "missing input #1");
    assert(mOp.mInputs[2] && "missing input #2");

    // Find the correct kernel type
    auto kernelFunc = Registrar<AddImplForward_ref_cpp<3>>::create({
        mOp.mInputs[0]->dataType(),
        mOp.mInputs[1]->dataType(),
        mOp.mInputs[2]->dataType(),
        mOp.mOutput->dataType()});

    // Call kernel
    kernelFunc(std::static_pointer_cast<Tensor>(mOp.mInputs[0])->size(),
        mOp.mInputs[0]->getImpl()->rawPtr(),
        mOp.mInputs[1]->getImpl()->rawPtr(),
        mOp.mInputs[2]->getImpl()->rawPtr(),
        mOp.mOutput->getImpl()->rawPtr());

    // Update producer-consumer data
    for (IOIndex_t inputIdx = 0; static_cast<NbElts_t>(inputIdx) < mNbConsumedData.size(); ++inputIdx)
        mNbConsumedData[inputIdx]+= getNbRequiredData(inputIdx); // each input is consumed by the minimum amount for a forward pass

    mNbProducedData[0]+= getRequiredMemory(0, {});
}

void Aidge::AddImpl_ref_cpp<3>::backward() {
    printf("Not implemented yet.\n");
}